/*
 usage: pcap2snoopy < pcapdump > snoopydump
 specify -v for some verbosity
*/

#include <u.h>
#include <libc.h>
#include <bio.h>

#define MIN(a, b)	((a) < (b) ? (a) : (b))

#define TCPDUMP_MAGIC_BE	"\xa1\xb2\xc3\xd4"
#define TCPDUMP_MAGIC_LE	"\xd4\xc3\xb2\xa1"

enum {
	PCAP_VERSION_MAJOR	=	2,
	PCAP_VERSION_MINOR	=	4,
};

int verbose = 0;

uvlong (*get64)(char *);
ulong (*get32)(char *);
ushort (*get16)(char *);

#define put16(p, v)	beput16(p, v)
#define put32(p, v)	beput32(p, v)
#define put64(p, v)	beput64(p, v)

ushort
leget16(char *sp)
{
	uchar *p = (uchar *)sp;
	return (ushort)p[0] | (ushort)p[1]<<8;
}

ulong
leget32(char *sp)
{
	return (ulong)leget16(sp) | (ulong)leget16(sp+2)<<16;
}

uvlong
leget64(char *sp)
{
	return (uvlong)leget32(sp) | (uvlong)leget32(sp+4)<<32;
}

ushort
beget16(char *sp)
{
	uchar *p = (uchar *)sp;
	return (ushort)p[0]<<8 | (ushort)p[1];
}

ulong
beget32(char *sp)
{
	return (ulong)beget16(sp)<<16 | (ulong)beget16(sp+2);
}

uvlong
beget64(char *sp)
{
	return (uvlong)beget32(sp)<<32 | (uvlong)beget32(sp+4);
}

void
beput16(char *sp, ushort v)
{
	uchar *p = (uchar *)sp;
	p[0] = (uchar)(v>>8);
	p[1] = (uchar)v;
}

void
beput32(char *sp, ulong v)
{
	beput16(sp, (ushort)(v>>16));
	beput16(sp+2, (ushort)v);
}

void
beput64(char *sp, uvlong v)
{
	beput32(sp, (ulong)(v>>32));
	beput32(sp+4, (ulong)v);
}


void
main(int argc, char *argv[])
{
	Biobuf b;
	char hdr[24];
	char inpkthdr[16];
	char outpkthdr[10];
	char *p;
	uvlong timestamp;
	ulong len, caplen;
	char buf[1024*64];
	long n, need;
	ulong maxlen, linktype;

	ARGBEGIN{
	case 'v':
		verbose = 1;
		break;
	default:
		fprint(2, "usage: pcap2snoopy [-v]\n");
		exits("usage");
	}ARGEND

	if(Binit(&b, 0, OREAD) < 0)
		sysfatal("Binit: %r");

	if(Bread(&b, hdr, sizeof hdr) != sizeof hdr)
		sysfatal("reading header: %r");

	p = hdr;
	if(memcmp(p, TCPDUMP_MAGIC_BE, 4) == 0){
		if(verbose)
			fprint(2, "big endian\n");
		get64 = beget64;
		get32 = beget32;
		get16 = beget16;
	}else if(memcmp(p, TCPDUMP_MAGIC_LE, 4) == 0){
		if(verbose)
			fprint(2, "little endian\n");
		get64 = leget64;
		get32 = leget32;
		get16 = leget16;
	}else{
		sysfatal("invalid magic (0x%ux%ux%ux%ux)", p[0], p[1], p[2], p[3]);
	}
	p += 4;

	if(get16(p) != PCAP_VERSION_MAJOR)
		sysfatal("wrong pcap major version (%d, need %d)", get16(p), PCAP_VERSION_MAJOR);
	p += 2;
	if(get16(p) != PCAP_VERSION_MINOR)
		fprint(2, "different minor version number (%d, only know of %d), continuing...\n", get16(p), PCAP_VERSION_MINOR);
	p += 2;

	/* 4 bytes long, gmt to local correction */
	p += 4;
	/* 4 bytes ulong, accuracy of timestamps */
	p += 4;
	maxlen = get32(p);
	p += 4;
	linktype = get32(p);
	p += 4;
	assert(p == hdr+sizeof hdr);

	if(verbose)
		fprint(2, "maxpacketlen=%uld linktype=0x%ulx\n", maxlen, linktype);

	for(;;){
		n = Bread(&b, inpkthdr, sizeof inpkthdr);
		if(n == 0)
			break;
		if(n < 0)
			sysfatal("reading packet header: %r");
		p = inpkthdr;
		timestamp = get32(p) * 1e9;
		timestamp += get32(p+4) * 1000;
		p += 8;
		caplen = get32(p);
		p += 4;
		len = get32(p);
		p += 4;
		USED(p);
		if(caplen != len)
			fprint(2, "packet not entirely captured (have %uld of %uld bytes)\n", caplen, len);

		p = outpkthdr;
		put16(p, caplen);
		p += 2;
		put64(p, timestamp);
		p += 8;
		USED(p);
		if(write(1, outpkthdr, sizeof outpkthdr) != sizeof outpkthdr)
			sysfatal("writing header: %r");

		need = caplen;
		while(need > 0){
			n = MIN(need, sizeof buf);
			if(Bread(&b, buf, n) != n)
				sysfatal("reading packet: %r\n");
			if(write(1, buf, n) != n)
				sysfatal("writing packet: %r");
			need -= n;
		}

		if(verbose)
			fprint(2, "have packet of %lud bytes\n", caplen);
	}
}
